@@ -1,20 +1,18 @@
 # Python SCALE Codec Library
 #
-# Copyright 2018-2020 openAware BV (NL).
-# This file is part of Polkascan.
+# Copyright 2018-2020 Stichting Polkascan (Polkascan Foundation).
 #
-# Polkascan is free software: you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation, either version 3 of the License, or
-# (at your option) any later version.
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
 #
-# Polkascan is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
+# http://www.apache.org/licenses/LICENSE-2.0
 #
-# You should have received a copy of the GNU General Public License
-# along with Polkascan. If not, see <http://www.gnu.org/licenses/>.
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
 
 import re
 from abc import ABC, abstractmethod
@@ -26,24 +24,71 @@ class Singleton(type):
     _instances = {}
 
     def __call__(cls, *args, **kwargs):
-        if cls not in cls._instances:
-            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
-        return cls._instances[cls]
+
+        if 'config_id' in kwargs:
+            instance_key = kwargs['config_id']
+        else:
+            instance_key = cls
+
+        if instance_key not in cls._instances:
+            cls._instances[instance_key] = super(Singleton, cls).__call__(*args, **kwargs)
+        return cls._instances[instance_key]
 
 
-class RuntimeConfiguration(metaclass=Singleton):
+class RuntimeConfigurationObject:
 
     @classmethod
     def all_subclasses(cls, class_):
         return set(class_.__subclasses__()).union(
             [s for c in class_.__subclasses__() for s in cls.all_subclasses(c)])
 
-    def __init__(self):
+    def __init__(self, config_id=None, ss58_format=None):
+        self.config_id = config_id
         self.type_registry = {}
+        self.__initial_state = False
         self.clear_type_registry()
         self.active_spec_version_id = None
+        self.chain_id = None
 
-    def get_decoder_class(self, type_string, spec_version_id='default'):
+        self.ss58_format = ss58_format
+
+    @classmethod
+    def convert_type_string(cls, name):
+
+        name = re.sub(r'T::', "", name, flags=re.IGNORECASE)
+        name = re.sub(r'<T>', "", name, flags=re.IGNORECASE)
+        name = re.sub(r'<T as Trait>::', "", name, flags=re.IGNORECASE)
+        name = re.sub(r'<T as Trait<I>>::', "", name, flags=re.IGNORECASE)
+        name = re.sub(r'<T as Config>::', "", name, flags=re.IGNORECASE)
+        name = re.sub(r'<T as Config<I>>::', "", name, flags=re.IGNORECASE)
+        name = re.sub(r'\n', "", name)
+        name = re.sub(r'(grandpa|session|slashing|limits|beefy_primitives|opaque)::', "", name)
+        name = re.sub(r'VecDeque<', "Vec<", name, flags=re.IGNORECASE)
+        name = re.sub(r'^Box<(.+)>$', r'\1', name, flags=re.IGNORECASE)
+
+        if name == '()':
+            return "Null"
+        if name.lower() in ['vec<u8>', '&[u8]']:
+            return "Bytes"
+        if name.lower() == '<lookup as staticlookup>::source':
+            return 'LookupSource'
+        if name.lower() == '<balance as hascompact>::type':
+            return 'Compact<Balance>'
+        if name.lower() == '<blocknumber as hascompact>::type':
+            return 'Compact<BlockNumber>'
+        if name.lower() == '<moment as hascompact>::type':
+            return 'Compact<Moment>'
+        if name.lower() == '<inherentofflinereport as inherentofflinereport>::inherent':
+            return 'InherentOfflineReport'
+
+        return name
+
+    def get_decoder_class(self, type_string: str, spec_version_id='default'):
+
+        if type_string.strip() == '':
+            return None
+
+        type_string = self.convert_type_string(type_string)
 
         decoder_class = self.type_registry.get('types', {}).get(type_string.lower(), None)
 
@@ -67,7 +112,7 @@ class RuntimeConfiguration(metaclass=Singleton):
             # Custom tuples
             elif type_string != '()' and type_string[0] == '(' and type_string[-1] == ')':
 
-                decoder_class = type(type_string, (RuntimeConfiguration().get_decoder_class('struct'),), {
+                decoder_class = type(type_string, (self.get_decoder_class('tuple'),), {
                     'type_string': type_string
                 })
 
@@ -81,21 +126,38 @@ class RuntimeConfiguration(metaclass=Singleton):
 
                 if type_parts:
                     # Create dynamic class for e.g. [u8; 4] resulting in array of u8 with 4 elements
-                    decoder_class = type(type_string, (RuntimeConfiguration().get_decoder_class('FixedLengthArray'),), {
+                    decoder_class = type(type_string, (self.get_decoder_class('FixedLengthArray'),), {
                         'sub_type': type_parts[0],
                         'element_count': int(type_parts[1])
                     })
 
+        if decoder_class:
+            # Attach RuntimeConfigurationObject to new class
+            decoder_class.runtime_config = self
+
         return decoder_class
 
     def clear_type_registry(self):
-        self.type_registry = {'types': {cls.type_string.lower(): cls for cls in self.all_subclasses(ScaleDecoder) if
-                                        cls.type_string}}
-        self.type_registry['types'].update({cls.__name__.lower(): cls for cls in self.all_subclasses(ScaleDecoder)})
+
+        if not self.__initial_state:
+            self.type_registry = {'types': {cls.type_string.lower(): cls for cls in self.all_subclasses(ScaleDecoder) if
+                                            cls.type_string}}
+
+            # Class names that contains '<' are excluded because of a side effect that is introduced in
+            # get_decoder_class: "Create dynamic class for Part1<Part2> based on Part1 and set class variable Part2 as
+            # sub_type" which won't get reset because class definitions always remain globally
+
+            self.type_registry['types'].update(
+                {cls.__name__.lower(): cls for cls in self.all_subclasses(ScaleDecoder) if '<' not in cls.__name__}
+            )
+
+        self.__initial_state = True
 
     def update_type_registry_types(self, types_dict):
         from scalecodec.types import Enum, Struct, Set
 
+        self.__initial_state = False
+
         for type_string, decoder_class_data in types_dict.items():
 
             if type(decoder_class_data) == dict:
@@ -103,18 +165,33 @@ class RuntimeConfiguration(metaclass=Singleton):
                 # Create dynamic decoder class
                 if decoder_class_data['type'] == 'struct':
 
-                    decoder_class = type(type_string, (Struct,), {'type_mapping': decoder_class_data['type_mapping']})
+                    if decoder_class_data.get('base_class'):
+                        base_cls = self.get_decoder_class(decoder_class_data['base_class'])
+                    else:
+                        base_cls = Struct
+
+                    decoder_class = type(type_string, (base_cls,), {'type_mapping': decoder_class_data['type_mapping']})
 
                 elif decoder_class_data['type'] == 'enum':
 
-                    decoder_class = type(type_string, (Enum,), {
+                    if decoder_class_data.get('base_class'):
+                        base_cls = self.get_decoder_class(decoder_class_data['base_class'])
+                    else:
+                        base_cls = Enum
+
+                    decoder_class = type(type_string, (base_cls,), {
                         'value_list': decoder_class_data.get('value_list'),
                         'type_mapping': decoder_class_data.get('type_mapping')
                     })
 
                 elif decoder_class_data['type'] == 'set':
 
-                    decoder_class = type(type_string, (Set,), {
+                    if decoder_class_data.get('base_class'):
+                        base_cls = self.get_decoder_class(decoder_class_data['base_class'])
+                    else:
+                        base_cls = Set
+
+                    decoder_class = type(type_string, (base_cls,), {
                         'value_list': decoder_class_data.get('value_list'),
                         'value_type': decoder_class_data.get('value_type', 'u64')
                     })
@@ -133,6 +210,9 @@ class RuntimeConfiguration(metaclass=Singleton):
         # Set runtime ID if set
         self.active_spec_version_id = type_registry.get('runtime_id')
 
+        # Set chain ID if set
+        self.chain_id = type_registry.get('chain_id')
+
         # Set versioning
         if 'versioning' in type_registry:
             self.type_registry['versioning'] = type_registry.get('versioning')
@@ -191,6 +271,17 @@ class ScaleBytes:
     def __str__(self):
         return "0x{}".format(self.data.hex())
 
+    def __eq__(self, other):
+        if not hasattr(other, 'data'):
+            return False
+        return self.data == other.data
+
+    def __len__(self):
+        return len(self.data)
+
+    def __repr__(self):
+        return "<{}(data=0x{})>".format(self.__class__.__name__, self.data.hex())
+
     def __add__(self, data):
 
         if type(data) == ScaleBytes:
@@ -204,6 +295,9 @@ class ScaleBytes:
         if type(data) == bytearray:
             return ScaleBytes(self.data + data)
 
+    def to_hex(self):
+        return f'0x{self.data.hex()}'
+
 
 class ScaleDecoder(ABC):
 
@@ -218,7 +312,9 @@ class ScaleDecoder(ABC):
     PRIMITIVES = ('bool', 'u8', 'u16', 'u32', 'u64', 'u128', 'u256', 'i8', 'i16', 'i32', 'i64', 'i128', 'i256', 'h160',
                   'h256', 'h512', '[u8; 4]', '[u8; 4]', '[u8; 8]', '[u8; 12]', '[u8; 16]', '[u8; 32]', '&[u8]')
 
-    def __init__(self, data, sub_type=None):
+    runtime_config = None
+
+    def __init__(self, data, sub_type=None, runtime_config=None):
 
         if sub_type:
             self.sub_type = sub_type
@@ -229,9 +325,18 @@ class ScaleDecoder(ABC):
         if data:
             assert(type(data) == ScaleBytes)
 
+        if runtime_config:
+            self.runtime_config = runtime_config
+
+        if not self.runtime_config:
+            # if no runtime config is provided, fallback on singleton
+            self.runtime_config = RuntimeConfiguration()
+
         self.data = data
         self.raw_value = ''
         self.value = None
+        self.data_start_offset = None
+        self.data_end_offset = None
 
     @classmethod
     def build_type_mapping(cls):
@@ -248,10 +353,8 @@ class ScaleDecoder(ABC):
                 for sub_type in sub_types:
                     tuple_contents = tuple_contents.replace(sub_type, sub_type.replace(',', '|'))
 
-            n = 1
-            for struct_element in tuple_contents.split(','):
-                type_mapping += (('col{}'.format(n), struct_element.strip().replace('|', ',')),)
-                n += 1
+            for tuple_element in tuple_contents.split(','):
+                type_mapping += (tuple_element.strip().replace('|', ','),)
 
             cls.type_mapping = type_mapping
 
@@ -274,101 +377,108 @@ class ScaleDecoder(ABC):
         self.raw_value += data.hex()
         return data
 
+    def get_used_bytes(self):
+        return self.data.data[self.data_start_offset:self.data_end_offset]
+
     @abstractmethod
     def process(self):
         raise NotImplementedError
 
     def decode(self, check_remaining=True):
+        self.data_start_offset = self.data.offset
         self.value = self.process()
+        self.data_end_offset = self.data.offset
 
-        if check_remaining and self.data.offset < self.data.length:
-            raise RemainingScaleBytesNotEmptyException('Current offset: {} / length: {}'.format(self.data.offset, self.data.length))
+        if check_remaining and self.data.offset != self.data.length:
+            raise RemainingScaleBytesNotEmptyException(
+                f'Decoding "{self.type_string}" - Current offset: {self.data.offset} / length: {self.data.length}'
+            )
 
         if self.data.offset > self.data.length:
             raise RemainingScaleBytesNotEmptyException(
-                'No more bytes available (offset: {} / length: {})'.format(self.data.offset, self.data.length))
+                f'Decoding "{self.type_string}" - No more bytes available (needed: {self.data.offset} / total: {self.data.length})'
+            )
 
         return self.value
 
     def __str__(self):
-        return str(self.value) or ''
+        return str(self.serialize()) or ''
 
-    def encode(self, value):
-        self.value = value
-        self.data = self.process_encode(value)
+    def __repr__(self):
+        return "<{}(value={})>".format(self.__class__.__name__, self.serialize())
+
+    def encode(self, value=None):
+
+        if value is not None:
+            self.value = value
+
+        self.data = self.process_encode(self.value)
         return self.data
 
     def process_encode(self, value):
         raise NotImplementedError("Encoding not implemented for this ScaleType")
 
     @classmethod
-    def get_decoder_class(cls, type_string, data=None, **kwargs):
-        """
-        :param type_string:
-        :param data:
-        :param kwargs:
-        :return: ScaleType
+    def get_decoder_class(cls, type_string, data=None, runtime_config=None, **kwargs):
         """
 
-        type_string = cls.convert_type(type_string)
+        Parameters
+        ----------
+        type_string
+        data
+        runtime_config
+        kwargs
 
-        decoder_class = RuntimeConfiguration().get_decoder_class(
-            type_string.lower(),
+        Returns
+        -------
+        ScaleType
+        """
+
+        if not runtime_config:
+            runtime_config = RuntimeConfiguration()
+
+        decoder_class = runtime_config.get_decoder_class(
+            type_string,
             spec_version_id=kwargs.get('spec_version_id', 'default')
         )
         if decoder_class:
-            return decoder_class(data, **kwargs)
+            return decoder_class(data=data, runtime_config=runtime_config, **kwargs)
 
         raise NotImplementedError('Decoder class for "{}" not found'.format(type_string))
 
     # TODO rename to decode_type (confusing when encoding is introduced)
     def process_type(self, type_string, **kwargs):
-        obj = self.get_decoder_class(type_string, self.data, **kwargs)
+        obj = self.get_decoder_class(type_string, self.data, runtime_config=self.runtime_config, **kwargs)
         obj.decode(check_remaining=False)
         return obj
 
     def serialize(self):
         return self.value
 
-    # TODO convert to TYPE_ALIAS per class Address: TYPE_ALIAS = ('<Lookup as StaticLookup>::Source',)
     @classmethod
     def convert_type(cls, name):
+        return RuntimeConfigurationObject.convert_type_string(name)
 
-        name = re.sub(r'T::', "", name)
-        name = re.sub(r'<T>', "", name)
-        name = re.sub(r'<T as Trait>::', "", name)
-        name = re.sub(r'\n', "", name)
-        name = re.sub(r'(grandpa|session|slashing)::', "", name)
 
-        if name == '()':
-            return "Null"
-        if name in ['Vec<u8>', '&[u8]']:
-            return "Bytes"
-        if name == '<Lookup as StaticLookup>::Source':
-            return 'Address'
-        if name == 'Vec<<Lookup as StaticLookup>::Source>':
-            return 'Vec<Address>'
-        if name == '<Balance as HasCompact>::Type':
-            return 'Compact<Balance>'
-        if name == '<BlockNumber as HasCompact>::Type':
-            return 'Compact<BlockNumber>'
-        if name == '<Moment as HasCompact>::Type':
-            return 'Compact<Moment>'
-        if name == '<InherentOfflineReport as InherentOfflineReport>::Inherent':
-            return 'InherentOfflineReport'
-        if name == 'RawAddress':
-            return 'Address'
-
-        return name
+class RuntimeConfiguration(RuntimeConfigurationObject, metaclass=Singleton):
+    pass
 
 
-# TODO move type_string and sub_type behaviour to this sub class
 class ScaleType(ScaleDecoder, ABC):
 
-    def __init__(self, data=None, sub_type=None, metadata=None):
+    def __init__(self, data=None, sub_type=None, metadata=None, runtime_config=None):
+        """
+
+        Parameters
+        ----------
+        data: ScaleBytes
+        sub_type: str
+        metadata: MetadataDecoder
+        runtime_config: RuntimeConfigurationObject
+        """
         self.metadata = metadata
         if not data:
             data = ScaleBytes(bytearray())
-        super().__init__(data, sub_type)
+        super().__init__(data, sub_type, runtime_config=runtime_config)
 
 
